#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from grib_check.Assert import Eq, Fail, IsIn, IsMultipleOf, Le, Ne
from grib_check.Report import Report

from .GeneralChecks import GeneralChecks


class Lam(GeneralChecks):
    def __init__(self, lookup_table, check_limits=False, check_validity=True):
        super().__init__(lookup_table, check_limits=check_limits, check_validity=check_validity)
        self.logger = logging.getLogger(__class__.__name__)

    def _basic_checks(self, message, p) -> Report:
        report = Report("Lam Basic Checks")
        report.add(IsIn(message["hour"], [0, 3, 6, 9, 12, 15, 18, 21]))
        report.add(IsIn(message["productionStatusOfProcessedData"], [4, 5]))
        report.add(Le(message["endStep"], 30 * 24))
        report.add(IsMultipleOf(message["step"], 3))

        report.add(self._check_date(message, p))

        return super()._basic_checks(message, p).add(report)

    # not registered in the lookup table
    def _statistical_process(self, message, p) -> Report:
        report = Report("Lam Statistical Process")

        topd = message.get("typeOfProcessedData", int)

        if topd in [0, 1, 2]:  # Analysis, Forecast, Analysis and forecast products
            pass
        elif topd in [3, 4]:  # Control forecast products
            report.add(Eq(message["productDefinitionTemplateNumber"], 11))
        else:
            report.add(Fail(f"Unsupported typeOfProcessedData {topd}"))
            return report

        if message["indicatorOfUnitOfTimeRange"] == 10:  # three hours
            # Three hourly is OK
            pass
        else:
            report.add(Eq(message["indicatorOfUnitOfTimeRange"], 1))
            report.add(IsMultipleOf(message["forecastTime"], 3))

        report.add(Eq(message["timeIncrementBetweenSuccessiveFields"], 0))
        report.add(IsMultipleOf(message["endStep"], 3))  # Every three hours

        return super()._statistical_process(message, p).add(report)

    def _point_in_time(self, message, p) -> Report:
        report = Report("Lam Point In Time")
        topd = message.get("typeOfProcessedData", int)
        if topd in [0, 1]:  # Analysis, Forecast
            if message["productDefinitionTemplateNumber"] == 1:
                report.add(Ne(message["numberOfForecastsInEnsemble"], 0))
                report.add(
                    Le(
                        message["perturbationNumber"],
                        message["numberOfForecastsInEnsemble"],
                    )
                )
        elif topd == 2:  # Analysis and forecast products
            pass
        elif topd == 3:  # Control forecast products
            report.add(Eq(message["productDefinitionTemplateNumber"], 1))
        elif topd == 4:  # Perturbed forecast products
            report.add(Eq(message["productDefinitionTemplateNumber"], 1))
            report.add(
                Le(
                    message["perturbationNumber"],
                    message["numberOfForecastsInEnsemble"],
                )
            )
        else:
            report.add(Fail(f"Unsupported typeOfProcessedData {topd}"))

        if message["indicatorOfUnitOfTimeRange"] == 10:
            # Three hourly is OK
            pass
        else:
            report.add(Eq(message["indicatorOfUnitOfTimeRange"], 1))
            report.add(IsMultipleOf(message["forecastTime"], 3))

        return super()._point_in_time(message, p).add(report)



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

from grib_check.checker.Tigge import Tigge
from grib_check.Grib import Grib
from grib_check.LookupTable import SimpleLookupTable

src_path = f"{os.path.dirname(os.path.realpath(__file__))}/../src/grib_check"


class TestTigge:
    def test_tigge_param_10v_good(self):
        tigge_params = (f"{src_path}/checker/TiggeParameters.jsonnet")
        checker = Tigge(SimpleLookupTable(tigge_params), check_limits=False, check_validity=False)
        grib = Grib("./tests/tigge/tigge_ecmf_sfc_10v.grib")
        message = next(grib)
        report = checker.validate(message)
        assert report.status() is True



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from grib_check.Assert import Eq, IsIn, IsMultipleOf, Le, Ne
from grib_check.Report import Report

from .GeneralChecks import GeneralChecks


class Wpmip(GeneralChecks):
    def __init__(self, lookup_table, check_limits=False, check_validity=True):
        super().__init__(lookup_table, check_limits=check_limits, check_validity=check_validity)
        self.logger = logging.getLogger(__class__.__name__)
        self.register_checks(
            {
                "pressure_level": self._pressure_level,
            }
        )

    def _basic_checks(self, message, p):
        report = Report("Wpmip Basic Checks")

        # WPMIP prod/test data
        report.add(IsIn(message["productionStatusOfProcessedData"], [16, 17]))

        # WPMIP centre/subCentre DGOV-577
        report.add(Eq(message["centre"], "323"))
        report.add(Ne(message["subCentre"], 0))

        # to use MARS new key "model"
        report.add(Le(message["backgroundProcess"], 255))
        report.add(Le(message["generatingProcessIdentifier"], 4))

        # CCSDS compression
        # https://codes.ecmwf.int/grib/format/grib2/ctables/5/0/
        report.add(Eq(message["dataRepresentationTemplateNumber"], 42))

        #       # Only 00, 06 12 and 18 Cycle OK
        #       report.add(IsIn(message["hour"], [0, 6, 12, 18]))

        report.add(Le(message["endStep"], 10 * 36))
        report.add(IsMultipleOf(message["step"], 6))
        report.add(Eq(message["bitsPerValue"], 16))
        report.add(self._check_date(message, p))

        return super()._basic_checks(message, p).add(report)
        # return report

    def _pressure_level(self, message, p):
        report = Report("WPMIP Pressure level")
        levels = [
            10,
            50,
            100,
            150,
            200,
            250,
            300,
            400,
            500,
            700,
            850,
            925,
            1000,
        ]
        report.add(IsIn(message["level"], levels))
        return report

    # not registered in the lookup table
    def _statistical_process(self, message, p) -> Report:
        report = Report("WPMIP Statistical Process")

        if message.get("indicatorOfUnitOfTimeRange") == 11:  # six hours
            # Six hourly is OK
            pass
        else:
            report.add(Eq(message["indicatorOfUnitOfTimeRange"], 1))
            report.add(IsMultipleOf(message["forecastTime"], 6))

        report.add(Eq(message["timeIncrementBetweenSuccessiveFields"], 0))
        report.add(IsMultipleOf(message["endStep"], 6))

        return super()._statistical_process(message, p).add(report)

    def _latlon_grid(self, message):
        report = Report(f"{__class__.__name__}.latlon_grid")

        report.add(Eq(message["Ni"], 1440))
        report.add(Eq(message["Nj"], 721))
        report.add(Eq(message["scanningMode"], 0))

        report.add(Eq(message["basicAngleOfTheInitialProductionDomain"], 0))
        # report.add(Missing(message, "subdivisionsOfBasicAngle"))
        report.add(Eq(message["latitudeOfFirstGridPoint"], 90000000))
        report.add(Eq(message["longitudeOfFirstGridPoint"], 0))
        report.add(Eq(message["latitudeOfLastGridPoint"], -90000000))
        report.add(Eq(message["longitudeOfLastGridPoint"], 359750000))
        report.add(Eq(message["iDirectionIncrement"], 250000))
        report.add(Eq(message["jDirectionIncrement"], 250000))

        return super()._latlon_grid(message).add(report)



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging

from grib_check.Assert import (
    Eq,
    Fail,
    IsIn,
    IsMultipleOf,
    Le,
    Ne,
    Pass,
)
from grib_check.KeyValue import KeyValue
from grib_check.Report import Report

from .Uerra import Uerra


class Crra(Uerra):
    def __init__(self, lookup_table, check_limits=False, check_validity=True):
        super().__init__(lookup_table, check_limits=check_limits, check_validity=check_validity)
        self.logger = logging.getLogger(__class__.__name__)

    def _basic_checks(self, message, p) -> Report:
        report = Report("CRRA Basic Checks")
        report.add(IsIn(message["productionStatusOfProcessedData"], [10, 11]))
        topd = message.get("typeOfProcessedData", int)
        report.add(IsIn(topd, [0, 1]))

        stream = message.get("stream", str)
        if stream != "moda":
            if topd == 0:
                report.add(Eq(message["step"], 0))
            else:
                report.add(IsIn(message["step"], [1, 2, 4, 5]) | IsMultipleOf(message["step"], 3))

        if message.get("paramId", int) not in [260651, 235072]:
            report.add(Eq(message["versionNumberOfGribLocalTables"], 0))

        return super()._basic_checks(message, p).add(report)

    def _point_in_time(self, message, p) -> Report:
        report = Report("CRRA Point in Time")

        topd = message.get("typeOfProcessedData", int)
        stream = message.get("stream", str)
        if topd in [0, 1]:  # Analysis, Forecast
            if stream == "dame" or stream == "moda":
                report.add(Eq(message["productDefinitionTemplateNumber"], 8))
            else:
                report.add(IsIn(message["productDefinitionTemplateNumber"], [0, 1]))
#       elif topd == 2:  # Analysis and forecast products
#           pass
        elif topd == 3:  # Control forecast products
            report.add(Eq(message["productDefinitionTemplateNumber"], 1))
        elif topd == 4:  # Perturbed forecast products
            # Is there always cf in tigge global datasets??
            report.add(
                Le(
                    message["perturbationNumber"],
                    message["numberOfForecastsInEnsemble"] - 1,
                )
            )
        else:
            report.add(Fail(f"Unsupported typeOfProcessedData {topd}"))

        if message["indicatorOfUnitOfTimeRange"] == 1:  # hourly
            report.add(
                IsIn(message["forecastTime"], [1, 2, 4, 5])
                | IsMultipleOf(message["forecastTime"], 3)
            )

        return report

    def _from_start(self, message, p):
        report = Report("CRRA From Start")
        stream = message.get("stream", str)
        if stream != "moda" and stream != "dame":
            report.add(Eq(message["startStep"], 0))
        report.add(self._statistical_process(message, p))

        endStep = message["endStep"]
        if endStep == 0:
            min_value, max_value = message.minmax()
            if min_value == 0 and max_value == 0:
                report.add(Pass(f"min and max are both {KeyValue(None, 0)} for {endStep}"))
            else:
                report.add(Fail(f"min and max should both be {KeyValue(None, 0)} for {endStep} but are {KeyValue(None, min_value)} and {KeyValue(None, max_value)}"))

        return report

    def _statistical_process(self, message, p) -> Report:
        report = Report("CRRA Statistical Process")

        topd = message.get("typeOfProcessedData", int)
        if topd.value() in [0, 1]:  # Analysis, Forecast
            report.add(Eq(message["productDefinitionTemplateNumber"], 8, f"topd={topd}"))
        else:
            report.add(Fail(f"Unsupported typeOfProcessedData {topd}"))
            return report

        report.add(Eq(message["numberOfMissingInStatisticalProcess"], 0))
        # report.add(Eq(message["indicatorOfUnitOfTimeForTheIncrementBetweenTheSuccessiveFieldsUsed"], 255))
        report.add(Eq(message["minuteOfEndOfOverallTimeInterval"], 0))
        report.add(Eq(message["secondOfEndOfOverallTimeInterval"], 0))

        return report

    def _check_validity_datetime(self, message):

        report = Report("CRRA Check Validity Datetime")

        stepType = message.get("stepType", str)
        stream = message.get("stream", str)
        topd = message.get("typeOfProcessedData", int)

        if Ne(stepType, "instant"):  # not instantaneous
            # Check only applies to accumulated, max etc.
            stepRange = message.get("stepRange", str)

            if Eq(stream, "dame") or Eq(stream, "moda"):
                year = message["year"].value()
                month = message["month"].value()
                day = message["day"].value()
                if month == 11:
                    month2 = 1
                    year2 = year + 1
                elif month == 12:
                    month2 = 2
                    year2 = year + 1
                else:
                    month2 = month + 2
                    year2 = year

                same_day = int(str(datetime.date(year, month, day)).replace('-', ''))
                next_day1 = datetime.date(year, month, day) + datetime.timedelta(days=1)
                next_day1 = int(str(next_day1).replace('-', ''))
                next_day2 = datetime.date(year, month, day) + datetime.timedelta(days=2)
                next_day2 = int(str(next_day2).replace('-', ''))
                last_date_month0 = datetime.date(year + int(month / 12), (month % 12) + 1, 1) - datetime.timedelta(days=1)
                last_date_month0 = int(str(last_date_month0).replace('-', ''))
                last_date_month1 = datetime.date(year2, month2, 1) - datetime.timedelta(days=1)
                last_date_month1 = int(str(last_date_month1).replace('-', ''))
                first_date_month1 = datetime.date(year + int(month / 12), (month % 12) + 1, 1)
                first_date_month1 = int(str(first_date_month1).replace('-', ''))
                first_date_month2 = datetime.date(year2, month2, 1)
                first_date_month2 = int(str(first_date_month2).replace('-', ''))

                # numberOfTimeRanges = message["numberOfTimeRanges"]
                lengthOfTimeRanges = [KeyValue("lengthOfTimeRange", v) for v in message.get_long_array("lengthOfTimeRange")]
                typeOfStatisticalProcessings = [KeyValue("typeOfStatisticalProcessing", v) for v in message.get_long_array("typeOfStatisticalProcessing")]

                typeOfTimeIncrements = [KeyValue("typeOfTimeIncrement", v) for v in message.get_long_array("typeOfTimeIncrement")]
                indicatorOfUnitForTimeRanges = [KeyValue("indicatorOfUnitForTimeRange", v) for v in message.get_long_array("indicatorOfUnitForTimeRange")]
                lengthOfTimeRanges = [KeyValue("lengthOfTimeRange", v) for v in message.get_long_array("lengthOfTimeRange")]
                indicatorOfUnitForTimeIncrements = [KeyValue("indicatorOfUnitForTimeIncrement", v) for v in message.get_long_array("indicatorOfUnitForTimeIncrement")]
                timeIncrements = [KeyValue("timeIncrement", v) for v in message.get_long_array("timeIncrement")]

                validityDateBefore = message["validityDate"]
                validityTimeBefore = message["validityTime"]

            # monthly/daily averages are archived under instant paramIds as param-db was not ready for all time-mean proper ones..
            # https://confluence.ecmwf.int/display/DGOV/Support+page+for+DGOV-399+CARRA+daily+and+monthly+GRIB+headers
            if Eq(stream, "dame"):

                report = Report("CRRA Check Validity Datetime - daily means")
                if typeOfStatisticalProcessings == [0]:
                    report = Report("dame - daily_mean_an/fc")
                    dame_validityDate = same_day
                    dame_validityTime = 21
                    if topd == 0:
                        dame_validityDate = same_day
                    elif topd == 1:
                        dame_validityDate = next_day1
                    [report.add(Eq(KeyValue("typeOfTimeIncrement", typeOfTimeIncrements[0]), 1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeRange", indicatorOfUnitForTimeRanges[0]), 1))]
                    [report.add(Eq(KeyValue("lengthOfTimeRange", lengthOfTimeRanges[0]), 21))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeIncrement", indicatorOfUnitForTimeIncrements[0]), 1))]
                    [report.add(Eq(KeyValue("timeIncrement", timeIncrements[0]), 3))]
                elif typeOfStatisticalProcessings == [1, 1]:
                    report = Report("dame - daily_sum_an/fc")
                    dame_validityDate = next_day2
                    dame_validityTime = 0
                    [report.add(Eq(KeyValue("typeOfTimeIncrement", typeOfTimeIncrements[0]), 1))]
                    [report.add(Eq(KeyValue("typeOfTimeIncrement", typeOfTimeIncrements[1]), 2))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeRange", indicatorOfUnitForTimeRanges[0]), 1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeRange", indicatorOfUnitForTimeRanges[1]), 1))]
                    [report.add(Eq(KeyValue("lengthOfTimeRange", lengthOfTimeRanges[0]), 24))]
                    [report.add(Eq(KeyValue("lengthOfTimeRange", lengthOfTimeRanges[1]), 12))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeIncrement", indicatorOfUnitForTimeIncrements[0]), 1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeIncrement", indicatorOfUnitForTimeIncrements[1]), 255))]
                    [report.add(Eq(KeyValue("timeIncrement", timeIncrements[0]), 12))]
                    [report.add(Eq(KeyValue("timeIncrement", timeIncrements[1]), 0))]
                elif typeOfStatisticalProcessings == [2, 2] or typeOfStatisticalProcessings == [3, 3]:
                    report = Report("dame - daily_min/max_an/fc")
                    dame_validityDate = next_day1
                    dame_validityTime = 0
                    [report.add(Eq(KeyValue("typeOfTimeIncrement", typeOfTimeIncrements[0]), 1))]
                    [report.add(Eq(KeyValue("typeOfTimeIncrement", typeOfTimeIncrements[1]), 2))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeRange", indicatorOfUnitForTimeRanges[0]), 1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeRange", indicatorOfUnitForTimeRanges[1]), 1))]
                    [report.add(Eq(KeyValue("lengthOfTimeRange", lengthOfTimeRanges[0]), 21))]
                    [report.add(Eq(KeyValue("lengthOfTimeRange", lengthOfTimeRanges[1]), 3))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeIncrement", indicatorOfUnitForTimeIncrements[0]), 1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeIncrement", indicatorOfUnitForTimeIncrements[1]), 255))]
                    [report.add(Eq(KeyValue("timeIncrement", timeIncrements[0]), 3))]
                    [report.add(Eq(KeyValue("timeIncrement", timeIncrements[1]), 0))]
                else:
                    report.add(Fail(f"Unsupported parameter in stream={stream}"))

                report.add(Eq(validityTimeBefore/100, dame_validityTime))
                report.add(Eq(validityDateBefore, dame_validityDate))

            elif Eq(stream, "moda"):

                report = Report("CRRA Check Validity Datetime - monthly means")

                moda_lotr1 = [669, 693, 717, 741]
                moda_lotr2 = [672, 696, 720, 744]

                if typeOfStatisticalProcessings == [0]:
                    report = Report("moda - monthly_mean_an/fc")
                    if topd == 0:
                        moda_validityDate = last_date_month0
                    elif topd == 1:
                        moda_validityDate = last_date_month1
                    moda_validityTime = 21
                    [report.add(Eq(KeyValue("typeOfTimeIncrement", typeOfTimeIncrements[0]), 1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeRange", indicatorOfUnitForTimeRanges[0]), 1))]
                    [report.add(IsIn(KeyValue("lengthOfTimeRange", lengthOfTimeRanges[0]), moda_lotr1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeIncrement", indicatorOfUnitForTimeIncrements[0]), 1))]
                    [report.add(Eq(KeyValue("timeIncrement", timeIncrements[0]), 3))]
                elif typeOfStatisticalProcessings == [2, 2] or typeOfStatisticalProcessings == [3, 3]:
                    report = Report("moda - monthly_min/max_an/fc")
                    moda_validityDate = first_date_month1
                    moda_validityTime = 0
                    [report.add(Eq(KeyValue("typeOfTimeIncrement", typeOfTimeIncrements[0]), 1))]
                    [report.add(Eq(KeyValue("typeOfTimeIncrement", typeOfTimeIncrements[1]), 2))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeRange", indicatorOfUnitForTimeRanges[0]), 1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeRange", indicatorOfUnitForTimeRanges[1]), 1))]
                    [report.add(IsIn(KeyValue("lengthOfTimeRange", lengthOfTimeRanges[0]), moda_lotr1))]
                    [report.add(Eq(KeyValue("lengthOfTimeRange", lengthOfTimeRanges[1]), 3))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeIncrement", indicatorOfUnitForTimeIncrements[0]), 1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeIncrement", indicatorOfUnitForTimeIncrements[1]), 255))]
                    [report.add(Eq(KeyValue("timeIncrement", timeIncrements[0]), 3))]
                    [report.add(Eq(KeyValue("timeIncrement", timeIncrements[1]), 0))]
                elif typeOfStatisticalProcessings == [0, 1, 1]:
                    report = Report("moda - monthly_daysum_an/fc")
                    moda_validityDate = first_date_month2
                    moda_validityTime = 0
                    [report.add(Eq(KeyValue("typeOfTimeIncrement", typeOfTimeIncrements[0]), 1))]
                    [report.add(Eq(KeyValue("typeOfTimeIncrement", typeOfTimeIncrements[1]), 1))]
                    [report.add(Eq(KeyValue("typeOfTimeIncrement", typeOfTimeIncrements[2]), 2))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeRange", indicatorOfUnitForTimeRanges[0]), 1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeRange", indicatorOfUnitForTimeRanges[1]), 1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeRange", indicatorOfUnitForTimeRanges[2]), 1))]
                    [report.add(IsIn(KeyValue("lengthOfTimeRange", lengthOfTimeRanges[0]), moda_lotr2))]
                    [report.add(Eq(KeyValue("lengthOfTimeRange", lengthOfTimeRanges[1]), 24))]
                    [report.add(Eq(KeyValue("lengthOfTimeRange", lengthOfTimeRanges[2]), 12))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeIncrement", indicatorOfUnitForTimeIncrements[0]), 1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeIncrement", indicatorOfUnitForTimeIncrements[1]), 1))]
                    [report.add(Eq(KeyValue("indicatorOfUnitForTimeIncrement", indicatorOfUnitForTimeIncrements[2]), 255))]
                    [report.add(Eq(KeyValue("timeIncrement", timeIncrements[0]), 24))]
                    [report.add(Eq(KeyValue("timeIncrement", timeIncrements[1]), 12))]
                    [report.add(Eq(KeyValue("timeIncrement", timeIncrements[2]), 0))]
                else:
                    report.add(Fail(f"Unsupported parameter in stream={stream}"))

                report.add(Eq(validityTimeBefore/100, moda_validityTime))
                report.add(Eq(validityDateBefore, moda_validityDate))

            else:

                # If we just set the stepRange (for non-instantaneous fields) to its
                # current value, then this causes the validity date and validity time
                # keys to be correctly computed.
                # Then we can compare the previous (possibly wrongly coded) value with
                # the newly computed one

                message.set("stepRange", stepRange)
                validityDate = message["validityDate"]
                validityTime = message["validityTime"]
                report.add(Eq(validityDate, validityDateBefore, f'Set stepRange={stepRange} has no effect on validityDate'))
                report.add(Eq(validityTime, validityTimeBefore, f'Set stepRange={stepRange} has no effect on validityTime'))

        return report

    def _pressure_level(self, message, p) -> Report:
        report = Report("CRRA Pressure Level")
        levels = [
            1000,
            975,
            950,
            925,
            900,
            875,
            850,
            825,
            800,
            750,
            700,
            600,
            500,
            400,
            300,
            250,
            200,
            150,
            100,
            70,
            50,
            30,
            20,
            10,
            7,
            5,
            3,
            2,
            1,
        ]
        report.add(IsIn(message["level"], levels, "valid pressure level"))
        return report

    def _height_level(self, message, p) -> Report:
        report = Report("CRRA Height Level")
        levels = [15, 30, 50, 75, 100, 150, 200, 250, 300, 400, 500, 750, 1000, 1250, 1500, 2000, 2500, 3000]
        report.add(IsIn(message["level"], levels, "valid height level"))
        return report



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

# from grib_check.GribCheck import GribCheck


class Config:
    def __init__(self):
        self.verbosity = 0
        self.convention = "tigge"
        self.warnflg = False
        self.valueflg = False
        self.zeroflg = False
        self.good = None
        self.bad = None
        self.report_verbosity = 3
        self.parameters = None
        self.num_threads = 1
        self.color = False
        # self.path = ["./tests/tigge_small/tigge_all.grib2"]
        self.path = ["./tests/tigge_small/tigge_cf_ecmwf.grib2"]
        self.debug = True
        self.failed_only = False
        self.format = "tree"


# class TestGribCheck:
#     def test_grib_check(self):
#         config = Config()
#         grib_check = GribCheck(config)
#         grib_check.run()



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from grib_check.Assert import Eq, EqDbl, Fail, Ge, Gt, IsIn, IsMultipleOf, Le, Lt, Ne
from grib_check.Report import Report

from .GeneralChecks import GeneralChecks


class S2SRefcst(GeneralChecks):
    def __init__(self, lookup_table, check_limits=False, check_validity=True):
        super().__init__(lookup_table, check_limits=check_limits, check_validity=check_validity)
        self.logger = logging.getLogger(__class__.__name__)
        self.register_checks(
            {
                "pressure_level": self._pressure_level,
            }
        )

    def _basic_checks(self, message, p) -> Report:
        report = Report("S2SRefcst Basic Checks")
        # Only 00, 06 12 and 18 Cycle OK
        report.add(IsIn(message["hour"], [0, 6, 12, 18]))
        report.add(
            IsIn(message["productionStatusOfProcessedData"], [4, 5])
        )  # TIGGE prod or test
        report.add(Le(message["endStep"], 30 * 24))
        report.add(IsMultipleOf(message["step"], 6))

        report.add(self._check_date(message, p))

        report.add(Eq(message["versionNumberOfGribLocalTables"], 0))

        return super()._basic_checks(message, p).add(report)

    def _latlon_grid(self, message) -> Report:
        report = Report(f"{__class__.__name__}.latlon_grid")

        tolerance = 1.0 / 1000000.0  # angular tolerance for grib2: micro degrees
        meridian = message["numberOfPointsAlongAMeridian"]
        parallel = message["numberOfPointsAlongAParallel"]

        north = message["latitudeOfFirstGridPoint"]
        south = message["latitudeOfLastGridPoint"]
        west = message["longitudeOfFirstGridPoint"]
        east = message["longitudeOfLastGridPoint"]

        ns = message["jDirectionIncrement"]
        we = message["iDirectionIncrement"]

        dnorth = message.get("latitudeOfFirstGridPointInDegrees", float)
        dsouth = message.get("latitudeOfLastGridPointInDegrees", float)
        dwest = message.get("longitudeOfFirstGridPointInDegrees", float)
        deast = message.get("longitudeOfLastGridPointInDegrees", float)

        dns = message.get("jDirectionIncrementInDegrees", float)
        dwe = message.get("iDirectionIncrementInDegrees", float)

        report.add(Gt(north, south, "north > south"))
        report.add(Gt(east, west, "east > west"))

        # Check that the grid is symmetrical */
        report.add(Eq(north, -south, "north == -south"))
        report.add(EqDbl(dnorth, -dsouth, tolerance, "dnorth == -dsouth"))
        report.add(
            Eq(parallel, (east - west) / we + 1, "parallel == (east - west) / we + 1")
        )
        report.add(
            Lt(
                ((deast - dwest) / dwe + 1 - parallel).abs(),
                1e-10,
                "math.fabs((deast - dwest) / dwe + 1 - parallel) < 1e-10",
            )
        )
        report.add(
            Eq(
                meridian,
                (north - south) / ns + 1,
                "meridian == (north - south) / ns + 1",
            )
        )
        report.add(
            Lt(
                ((dnorth - dsouth) / dns + 1 - meridian).abs(),
                1e-10,
                "math.fabs((dnorth - dsouth) / dns + 1 - meridian) < 1e-10 ",
            )
        )

        # Check that the field is global */
        area = (dnorth - dsouth) * (deast - dwest)
        globe = 360.0 * 180.0
        report.add(Le(area, globe, "area <= globe"))
        report.add(Ge(area, globe * 0.95, "area >= globe*0.95"))

        return super()._latlon_grid(message).add(report)

    # not registered in the lookup table
    def _statistical_process(self, message, p) -> Report:
        report = Report("S2SRefcst Statistical Process")

        topd = message.get("typeOfProcessedData", int)

        if topd in [0, 1, 2]:  # Analysis, Forecast, Analysis and forecast products
            pass
        elif topd in [3, 4]:  # Control forecast products, Perturbed forecast products
            report.add(Eq(message["productDefinitionTemplateNumber"], 61))
        else:
            report.add(Fail(f"Unsupported typeOfProcessedData {topd}"))
            return report

        if message["indicatorOfUnitOfTimeRange"] == 11:  # six hours
            # Six hourly is OK
            pass
        else:
            report.add(Eq(message["indicatorOfUnitOfTimeRange"], 1))
            report.add(IsMultipleOf(message["forecastTime"], 6))

        report.add(Eq(message["timeIncrementBetweenSuccessiveFields"], 0))
        report.add(IsMultipleOf(message["step"], 6))

        return super()._statistical_process(message, p).add(report)

    def _point_in_time(self, message, p) -> Report:
        report = Report("S2SRefcst Point In Time")
        topd = message.get("typeOfProcessedData", int)
        if topd in [0, 1]:  # Analysis, Forecast
            if message["productDefinitionTemplateNumber"] == 1:
                report.add(
                    Ne(message["numberOfForecastsInEnsemble"], 0, f"topd={topd}")
                )
                report.add(
                    Le(
                        message["perturbationNumber"],
                        message["numberOfForecastsInEnsemble"],
                        f"topd={topd}",
                    )
                )
        elif topd == 2:  # Analysis and forecast products
            pass
        elif topd == 3:  # Control forecast products
            report.add(
                Eq(message["productDefinitionTemplateNumber"], 60, f"topd={topd}")
            )
        elif topd == 4:  # Perturbed forecast products
            report.add(
                Eq(message["productDefinitionTemplateNumber"], 60, f"topd={topd}")
            )
            report.add(
                Lt(
                    message["perturbationNumber"],
                    message["numberOfForecastsInEnsemble"],
                    f"topd={topd}",
                )
            )
        else:
            report.add(Fail(f"Unsupported typeOfProcessedData {topd}"))

        if message["indicatorOfUnitOfTimeRange"] == 11:
            # Six hourly is OK
            pass
        else:
            report.add(Eq(message["indicatorOfUnitOfTimeRange"], 1))
            report.add(IsMultipleOf(message["forecastTime"], 6))

        return super()._point_in_time(message, p).add(report)



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os


class FileScanner:
    def __init__(self, paths):
        self.fns = []
        self.logger = logging.getLogger(__class__.__name__)
        for path in paths:
            if os.path.isdir(path):
                self.logger.debug(f"Scanning {path}")
                for root, dirs, fns in os.walk(path):
                    for fn in fns:
                        self.fns.append(root + "/" + fn)
            else:
                self.fns.append(path)

    def __iter__(self):
        self.pos = 0
        return self

    def __next__(self) -> str:
        if self.pos < len(self.fns):
            fn = self.fns[self.pos]
            self.pos += 1
            return fn
        else:
            raise StopIteration



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from grib_check.KeyValue import KeyValue


class TestIndexedLookupTable:
    def test_init(self):
        a = KeyValue("a", 5)
        assert str(a) == "a(5)"

    def test_op(self):
        a = KeyValue("a", 5)
        b = KeyValue("b", 10)
        c = KeyValue("c", 15)

        x = a + b
        assert x.key() == "a(5) + b(10)"
        assert x.value() == 15

        x = a - b
        assert x.key() == "a(5) - b(10)"
        assert x.value() == -5

        x = c % 10
        assert x.key() == "c(15) % 10"
        assert x.value() == 5

        x = -a
        assert x.key() == "-(a(5))"
        assert x.value() == -5

    def test_parentheses(self):
        a = KeyValue("a", 5)
        b = KeyValue("b", 10)
        c = KeyValue("c", 15)

        x = a + b * c
        assert x.key() == "a(5) + b(10) * c(15)"
        assert x.value() == 5 + 10 * 15

        x = (a + b) * c
        assert x.key() == "(a(5) + b(10)) * c(15)"
        assert x.value() == (5 + 10) * 15

        x = (a + b) % 10
        assert x.key() == "(a(5) + b(10)) % 10"
        assert x.value() == (5 + 10) % 10

        x = -(a + b)
        assert x.key() == "-(a(5) + b(10))"
        assert x.value() == -(5 + 10)

    def test_key_suffix(self):
        a = KeyValue("a", 5, key_suffix="[0]")
        b = KeyValue("b", 10, key_suffix="[1]")
        c = KeyValue("c", 15, key_suffix="[2]")

        x = a + b * c
        assert x.key() == "a[0](5) + b[1](10) * c[2](15)"
        assert x.value() == 5 + 10 * 15



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import sys

from eccodes import (
    codes_get_gaussian_latitudes,
    codes_grib_new_from_file,
)

from .Assert import Fail
from .Message import Message
from .Report import Report


class Grib:
    def __init__(self, path):
        self.__report = Report(f"File: {path}")
        self.__position = 0
        self.__f = None
        try:
            self.__f = open(path, "rb")
        except Exception as e:
            print(f"{path}, {str(e)}", file=sys.stderr)
            self.__report.add(Fail(f"Could not open file: {path}"))
            return

    def __del__(self):
        if self.__f is not None:
            self.__f.close()

    def __iter__(self):
        self.__position = 0
        return self

    def __next__(self):
        try:
            handle = codes_grib_new_from_file(self.__f)
        except Exception:
            self.__report.add(Fail(f"Could not read message[{self.__position + 1}]"))
            handle = None

        if handle is not None:
            self.__position += 1
            return Message(handle=handle, position=self.__position)
        else:
            raise StopIteration

    def report(self):
        return self.__report


def get_gaussian_latitudes(order):
    return codes_get_gaussian_latitudes(order)



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging

# from Message import Message
import math

from .KeyValue import makeKV
from .TermColor import TermColor


class Assert:
    def __init__(self, comment=None):
        self._logger = logging.getLogger(__class__.__name__)
        self._status = True
        self._comment = comment
        raise NotImplementedError(
            "Assert class is abstract and should not be instantiated directly"
        )

    def as_string(self, color=False, comment_position="below") -> str:
        if comment_position == "right":
            comment = f" : ({self._comment})" if self._comment is not None else ""
        elif comment_position == "below":
            comment = f"\n{self._comment}" if self._comment is not None else ""
        else:
            raise ValueError("comment_position must be either 'right' or 'below'")

        if color:
            return (
                f"{self._as_string(color)}{TermColor.OKBLUE}{comment}{TermColor.ENDC}"
            )
        else:
            return f"{self._as_string(color)}{comment}"

    def __str__(self) -> str:
        return self.as_string(color=False, comment_position="right")

    def __or__(self, other):
        return Or(self, other)

    def __and__(self, other):
        return And(self, other)

    def __bool__(self) -> bool:
        return self._status

    def _as_string(self, color=False) -> str:
        raise NotImplementedError


class AssertTrue(Assert):
    def __init__(self, status, msg, comment=None):
        self._status = bool(status)
        self._comment = comment
        self.__msg = msg

    def _as_string(self, color=False) -> str:
        return f"{self.__msg}"


class And(Assert):
    def __init__(self, lhs: Assert, rhs: Assert, comment=None):
        self.__lsh = lhs
        self.__rsh = rhs
        self._comment = comment
        self._status = bool(self.__lsh and self.__rsh)

    def _as_string(self, color=False) -> str:
        if color:
            return f"{self.__lsh.as_string(color)} {TermColor.OKCYAN}and{TermColor.ENDC} {self.__rsh.as_string(color)}"
        else:
            return f"{self.__lsh.as_string(color)} and {self.__rsh.as_string(color)}"


class Or(Assert):
    def __init__(self, lhs: Assert, rhs: Assert, comment=None):
        self.__lhs = lhs
        self.__rhs = rhs
        self._status = bool(self.__lhs or self.__rhs)
        self._comment = comment

    def _as_string(self, color=False) -> str:
        if color:
            return f"{self.__lhs.as_string(color)} {TermColor.OKCYAN}or{TermColor.ENDC} {self.__rhs.as_string(color)}"
        else:
            return f"{self.__lhs.as_string(color)} or {self.__rhs.as_string(color)}"


class IsIn(Assert):
    def __init__(self, actual, expected, comment=None):
        self.__actual = makeKV(actual)
        self.__expected = makeKV(expected)
        self._status = self.__actual.value() in self.__expected.value()
        self._comment = comment

    def _as_string(self, color=False) -> str:
        if color:
            return f"{self.__actual} in {self.__expected}"
        else:
            return f"{self.__actual} in {self.__expected}"


class IsMultipleOf(Assert):
    def __init__(self, actual, multiplier, comment=None):
        self.__actual = makeKV(actual)
        self.__multiplier = multiplier
        self.__mod_value = self.__actual % self.__multiplier
        self._status = self.__mod_value == 0
        self._comment = comment

    def _as_string(self, color=False) -> str:
        if color:
            return f"{self.__mod_value} == {makeKV(0)}"
        else:
            return f"{self.__mod_value} == {makeKV(0)}"


class Missing(Assert):
    def __init__(self, message, key, comment=None):
        self.__key_is_missing = message[key].value() is None
        if not self.__key_is_missing:
            self.__value_is_missing = message.is_missing(key)
            self._status = self.__value_is_missing
        else:
            self.__value_is_missing = None
            self._status = False
        self.__key = key
        self._comment = comment

    def _as_string(self, color=False) -> str:
        return f"{self.__key} exists({not self.__key_is_missing}) and has a missing value({None if self.__value_is_missing is None else self.__value_is_missing})"


class Exists(Assert):
    def __init__(self, message, key, comment=None):
        self.__key_is_missing = message[key].value() is None
        if not self.__key_is_missing:
            self.__value_is_missing = message.is_missing(key)
            self._status = not self.__value_is_missing
        else:
            self.__value_is_missing = None
            self._status = False
        self.__key = key
        self._comment = comment

    def _as_string(self, color=False) -> str:
        return f"{self.__key} exists({not self.__key_is_missing}) and has a non-missing value({None if self.__value_is_missing is None else not self.__value_is_missing})"


class Eq(Assert):
    def __init__(self, lsh, rhs, comment=None):
        self.__lsh = makeKV(lsh)
        self.__rhs = makeKV(rhs)
        self._status = self.__lsh == self.__rhs
        self._comment = comment

    def _as_string(self, color=False) -> str:
        if color:
            return f"{self.__lsh} == {self.__rhs}"
        else:
            return f"{self.__lsh} == {self.__rhs}"


class EqDbl(Assert):
    def __init__(self, lsh, rhs, tolerance, comment=None):
        self.__lsh = makeKV(lsh)
        self.__rhs = makeKV(rhs)
        self.__tolerance = tolerance
        self._status = math.fabs((self.__lsh - self.__rhs).value()) <= self.__tolerance
        self._comment = comment

    def _as_string(self, color=False) -> str:
        if color:
            return f"{self.__lsh} == {self.__rhs} within {self.__tolerance}"
        else:
            return f"{self.__lsh} == {self.__rhs} within {self.__tolerance}"


class Ne(Assert):
    def __init__(self, lsh, rhs, comment=None):
        self.__lsh = makeKV(lsh)
        self.__rhs = makeKV(rhs)
        self._status = self.__lsh != self.__rhs
        self._comment = comment

    def _as_string(self, color=False) -> str:
        if color:
            return f"{self.__lsh} != {self.__rhs}"
        else:
            return f"{self.__lsh} != {self.__rhs}"


class Ge(Assert):
    def __init__(self, lsh, rhs, comment=None):
        self.__lsh = makeKV(lsh)
        self.__rhs = makeKV(rhs)
        try:
            self._status = self.__lsh >= self.__rhs
        except TypeError:
            self._status = False
        self._comment = comment

    def _as_string(self, color=False) -> str:
        if color:
            return f"{self.__lsh} >= {self.__rhs}"
        else:
            return f"{self.__lsh} >= {self.__rhs}"


class Le(Assert):
    def __init__(self, lsh, rhs, comment=None):
        self.__lsh = makeKV(lsh)
        self.__rhs = makeKV(rhs)
        try:
            self._status = self.__lsh <= self.__rhs
        except TypeError:
            self._status = False
        self._comment = comment

    def _as_string(self, color=False) -> str:
        if color:
            return f"{self.__lsh} <= {self.__rhs}"
        else:
            return f"{self.__lsh} <= {self.__rhs}"


class Gt(Assert):
    def __init__(self, lsh, rhs, comment=None):
        self.__lsh = makeKV(lsh)
        self.__rhs = makeKV(rhs)
        self._status = self.__lsh > self.__rhs
        self._comment = comment

    def _as_string(self, color=False) -> str:
        if color:
            return f"{self.__lsh} > {self.__rhs}"
        else:
            return f"{self.__lsh} > {self.__rhs}"


class Lt(Assert):
    def __init__(self, lsh, rhs, comment=None):
        self.__lsh = makeKV(lsh)
        self.__rhs = makeKV(rhs)
        self._status = self.__lsh < self.__rhs
        self._comment = comment

    def _as_string(self, color=False) -> str:
        if color:
            return f"{self.__lsh} < {self.__rhs}"
        else:
            return f"{self.__lsh} < {self.__rhs}"


class Fail(Assert):
    def __init__(self, msg, comment=None):
        self._status = False
        self.__msg = msg
        self._comment = comment

    def _as_string(self, color=False) -> str:
        return f"{self.__msg}"


class Pass(Assert):
    def __init__(self, msg, comment=None):
        self._comment = comment
        self.__msg = msg
        self._status = True

    def _as_string(self, color=False) -> str:
        return f"{self.__msg}"



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

from grib_check.checker.Crra import Crra
from grib_check.Grib import Grib
from grib_check.LookupTable import SimpleLookupTable

src_path = f"{os.path.dirname(os.path.realpath(__file__))}/../src/grib_check"


class TestCrra:
    def test_crra_param_ws_good(self):
        crra_params = (f"{src_path}/checker/CrraParameters.jsonnet")
        checker = Crra(SimpleLookupTable(crra_params), check_limits=False, check_validity=False)
        grib = Grib("./tests/crra/crra_an_no-ar-pa_pl_ws.grib")
        message = next(grib)
        report = checker.validate(message)
        assert report.status() is True



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import argparse
import os
import sys

from eccodes import (
    codes_get_long,
    codes_get_string,
    codes_grib_new_from_file,
    codes_release,
)

from grib_check.FileScanner import FileScanner


class TiggeName:
    def __init__(self, list_mode=False, compare_mode=False):
        self.filename = ""
        self.error = 0
        self.field = 0
        self.param = "unknown"

        self.list_mode = list_mode
        self.compare_mode = compare_mode

    def __get(self, h, what) -> int:
        val = -1
        try:
            val = codes_get_long(h, what)
        except Exception as e:
            print(
                "%s, field %d [%s]: cannot get %s: %s"
                % (self.filename, self.field, self.param, what, str(e))
            )
            self.error += 1
        return val

    def __sget(self, h, what) -> str:
        val = None
        try:
            val = codes_get_string(h, what)
        except Exception as e:
            print(
                "%s, field %d [%s]: cannot get %s: %s"
                % (self.filename, self.field, self.param, what, str(e))
            )
            self.error += 1
        return val

    def __verify(self, h, full, base):
        level = 0
        number = 0

        marstype = self.__sget(h, "type")
        levtype = self.__sget(h, "levtype")

        if marstype == "fc":
            number = self.__get(h, "number")

        if levtype == "sfc":
            levtype = "sl"
        else:
            level = self.__get(h, "level")

        wmo_name = "z_tigge_c_%s_%08ld%04ld00_%s_%s_%s_%s_%04ld_%03ld_%04ld_%s.grib" % (
            self.__sget(h, "origin"),
            0 if self.compare_mode else self.__get(h, "date"),
            0 if self.compare_mode else self.__get(h, "time"),
            self.__sget(h, "model"),
            "xxxx" if self.compare_mode else self.__sget(h, "expver"),
            marstype,
            levtype,
            self.__get(h, "step"),
            number,
            level,
            self.__sget(h, "tigge_short_name"),
        )

        if self.list_mode:
            print("%s" % wmo_name)
        elif base != wmo_name:
            print("WRONG FILE NAME:   %s\nCORRECT FILE NAME: %s" % (base, wmo_name))
            self.error += 1

    def validate(self, path):
        try:
            f = open(path, "rb")
        except Exception as e:
            print("%s: %s" % (path, str(e)))
            self.error += 1
            return

        err = 0
        count = 0

        self.filename = path
        self.field = 0

        while True:
            h = None
            try:
                h = codes_grib_new_from_file(f)
            except Exception as e:
                err += 1
                last_error_message = str(e)
            if h is None:
                break
            self.field += 1
            self.__verify(h, path, os.path.basename(path))
            codes_release(h)
            count += 1
            self.param = "unknown"

        f.close()

        if err != 0:
            print("%s: grib_handle_new_from_file: %s" % (path, last_error_message))
            self.error += 1
            return

        if count == 0:
            print("%s does not contain any GRIBs" % path)
            self.error += 1
            return

    def get_error_counter(self):
        return self.error


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l", "--list-mode", help="enable list mode", action="store_true"
    )
    parser.add_argument(
        "-c", "--compare-mode", help="enable compare mode", action="store_true"
    )
    parser.add_argument(
        "path", nargs="+", help="path to a GRIB file or directory", type=str
    )
    args = parser.parse_args()

    tigge_name = TiggeName(list_mode=args.list_mode, compare_mode=args.compare_mode)

    for filename in FileScanner(args.path):
        tigge_name.validate(filename)

    sys.exit(0 if tigge_name.get_error_counter() == 0 else 1)


if __name__ == "__main__":
    main()



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import numpy as np

from .Assert import Assert
from .TermColor import TermColor


class RWarning:
    def __init__(self, msg):
        self.__msg = msg

    def __str__(self):
        return self.__msg


class RError:
    def __init__(self, msg):
        self.__msg = msg

    def __str__(self):
        return self.__msg


class Report:
    def __init__(self, name=None):
        self.__entries = list()
        self.__logger = logging.getLogger(__name__)
        self.__pass_str = "PASS"
        self.__fail_str = "FAIL"
        self.__none_str = "----"
        self.__status = None
        self.__name = name

    def __as_string_tree(self, entries, level, max_level, color, failed_only):
        shift = 0
        output = ""
        if color:
            pass_str = f"{TermColor.OKGREEN}{self.__pass_str}{TermColor.ENDC}"
            fail_str = f"{TermColor.FAIL}{self.__fail_str}{TermColor.ENDC}"
            none_str = f"{TermColor.OKCYAN}{self.__none_str}{TermColor.ENDC}"
        else:
            pass_str = self.__pass_str
            fail_str = self.__fail_str
            none_str = self.__none_str

        if max_level is None or level <= max_level:
            if self.__name is not None:
                if self.__status is None:
                    if not failed_only:
                        output = "  " * level + f"{none_str}: {self.__name}\n"
                elif self.__status is True or self.__status is np.bool_(True):
                    if not failed_only:
                        output = "  " * level + f"{pass_str}: {self.__name}\n"
                elif self.__status is False or self.__status is np.bool_(False):
                    output = "  " * level + f"{fail_str}: {self.__name}\n"
                else:
                    print(f"self.__status={self.__status}, type={type(self.__status)}")
                    raise NotImplementedError

                shift = 1

            if max_level is None or level + shift <= max_level:
                for entry in entries:
                    if isinstance(entry, Report):
                        output += entry.__as_string_tree(
                            entry.__entries,
                            level + shift,
                            max_level,
                            color,
                            failed_only,
                        )
                    elif isinstance(entry, Assert):
                        msg = entry.as_string(color)
                        status = bool(entry)
                        if not failed_only or not status:
                            # output += "  " * (level + shift) + f'{pass_str if status else fail_str}: {msg}\n'
                            tmp = (
                                "  " * (level + shift)
                                + f"{pass_str if status else fail_str}: {msg}"
                            )
                            tmp = tmp.replace("\n", f'\n      {"  " * (level + shift)}')
                            output += f"{tmp}\n"
                    elif type(entry) is str:
                        if not failed_only:
                            output += "  " * (level + shift) + f"{entry}\n"
                    else:
                        raise NotImplementedError

        return output

    def __as_string_short(self, report, color, failed_only, path):
        output = ""
        sep = " <- "

        if color:
            pass_str = f"{TermColor.OKGREEN}{self.__pass_str}{TermColor.ENDC}"
            fail_str = f"{TermColor.FAIL}{self.__fail_str}{TermColor.ENDC}"
            none_str = f"{TermColor.OKCYAN}{self.__none_str}{TermColor.ENDC}"
        else:
            pass_str = self.__pass_str
            fail_str = self.__fail_str
            none_str = self.__none_str

        if report.__name is not None:
            if color:
                path += f"{TermColor.OKCYAN}{report.__name}{TermColor.ENDC}{TermColor.SEP}{sep}{TermColor.ENDC}"
            else:
                path += f"{report.__name}{sep}"

        for entry in report.__entries:
            if isinstance(entry, Report):
                output += self.__as_string_short(entry, color, failed_only, path)
            elif isinstance(entry, Assert):
                status = bool(entry)
                if not failed_only or not status:
                    if status is True:
                        if color:
                            output += (
                                f"{pass_str}: {path}{entry.as_string(color, 'right')}\n"
                            )
                        else:
                            output += (
                                f"{pass_str}: {path}{entry.as_string(color, 'right')}\n"
                            )
                    elif status is False:
                        if color:
                            output += (
                                f"{fail_str}: {path}{entry.as_string(color, 'right')}\n"
                            )
                        else:
                            output += (
                                f"{fail_str}: {path}{entry.as_string(color, 'right')}\n"
                            )
                    else:
                        raise NotImplementedError
                # output += f"{path}{entry.as_string(color, comment_position='right')}\n"
            elif type(entry) is str:
                if not failed_only:
                    output += f"{none_str}: {path}{entry}\n"
            else:
                raise NotImplementedError

        return output

    def rename(self, name):
        self.__name = name

    def rename_anonymous_report(self, name):
        if self.__name is None:
            self.__name = name

    def as_string(self, max_level=None, color=False, failed_only=False, output_type=None):
        if output_type == "short":
            return self.__as_string_short(self, color, failed_only, "")
        elif output_type == "tree":
            return self.__as_string_tree(
                self.__entries, 0, max_level, color, failed_only
            )
        else:
            raise NotImplementedError(f"Unknown format: {output_type}")

    def status(self):
        return self.__status

    def __str__(self):
        return self.__as_string_tree(
            entries=self.__entries,
            level=0,
            max_level=None,
            color=False,
            failed_only=False,
        )

    def add(self, entry):
        if not (
            isinstance(entry, Assert) or type(entry) is Report or type(entry) is str
        ):
            print(f"entry={entry}, type={type(entry)}")
        assert isinstance(entry, Assert) or type(entry) is Report or type(entry) is str
        if isinstance(entry, Assert):
            if self.__status is None:
                self.__status = bool(entry)
            else:
                if bool(entry) is False:
                    self.__status = False
        elif isinstance(entry, Report):
            if self.__status is None:
                self.__status = entry.status()
            elif entry.status() is False:
                self.__status = False
            else:
                if entry.status() is not None:
                    self.__status = self.__status and entry.status()
        elif type(entry) is str:
            pass

        assert type(self.__status) is bool or type(self.__status) is np.bool_ or self.__status is None, f"self.__status={self.__status}, type={type(self.__status)}"

        self.__entries.append(entry)

        return self

    def error(self, msg):
        # TODO: Implement error handling
        self.add(RError(msg))

    def warning(self, entry):
        # TODO: Implement warning handling
        self.add(RWarning(entry))



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from .Assert import Eq, Fail
from .LookupTable import LookupTable
from .Message import Message
from .Report import Report
from .Test import Test


class CheckEngine:
    class DefaultTest(Test):
        def __init__(self, message: Message, parameter: dict, check_map: dict):
            self.logger = logging.getLogger(__class__.__name__)
            assert parameter is not None
            assert message is not None
            assert check_map is not None

            self.__message = message
            self.__parameter = parameter
            self.__check_map = check_map

        def run(self) -> Report:
            data = self.__parameter
            report = Report(f"{data['name']}")
            for check_func in data["checks"]:
                try:
                    report.add(self.__check_map[check_func](self.__message, data))
                except KeyError:
                    report.add(Fail(f'Check function "{check_func}" not found'))
            return report

    def __init__(self, lookup_table: LookupTable):
        self.logger = logging.getLogger(__class__.__name__)
        assert lookup_table is not None
        self._test_store = lookup_table
        self._check_map = None

    def _create_test(self, message: Message, parameters: dict) -> Test:
        assert parameters is not None
        assert self._check_map is not None
        return self.DefaultTest(message, parameters, self._check_map)

    # def set_checks(self, check_map: dict):
    #     assert check_map is not None
    #     self._check_map = check_map

    def register_checks(self, check_funcs: dict):
        assert check_funcs is not None
        if self._check_map is None:
            self._check_map = dict()
        for name, func in check_funcs.items():
            if name in self._check_map:
                raise ValueError(f"Check function {name} already registered")
            self._check_map[name] = func

    def validate(self, message) -> Report:
        report = Report()
        kv, store_report = self._test_store.get_element(message)
        if kv is not None:
            test = self._create_test(message, kv)
            report.add(store_report)
            report.add(test.run())

            if "expected" in kv:
                expected_report = Report("Expected Values")
                for expected in kv["expected"]:
                    if "key" in expected and "value" in expected:
                        expected_report.add(
                            Eq(message[expected["key"]], expected["value"])
                        )
                report.add(expected_report)
        else:
            self.logger.debug(f"Could not find parameter for: {message}")
            report.add(Fail("Could not find parameter"))
            test_sub_report = Report()
            test_sub_report.add(message.get_report())
            report.add(test_sub_report)

        return report



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest
from eccodes import CODES_PRODUCT_GRIB, codes_new_from_samples

from grib_check.Message import Message


class TestMessage:
    def test_position(self):
        handle = codes_new_from_samples("GRIB2", product_kind=CODES_PRODUCT_GRIB)
        with pytest.raises(Exception):
            Message(handle=handle, position=0)

    # def test_type(self):
    #     handle = codes_new_from_samples("GRIB2", product_kind=CODES_PRODUCT_GRIB)
    #     msg = Message(handle=handle, position=1)
    #
    #     # Native type is str
    #     kv = msg.get("identifier")
    #     assert kv.type() is str
    #
    #     # Value can be converted to int
    #     kv = msg.get("identifier", datatype=int)
    #     assert kv.type() is int
    #
    #     # Value cannot be converted to float
    #     kv = msg.get("identifier", datatype=float)
    #     assert kv.type() is type(None)



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from grib_check.checker.GeneralChecks import GeneralChecks
from grib_check.Grib import Grib
from grib_check.LookupTable import SimpleLookupTable


class TestWmoChecker:
    def test_load_data_from_file(self):
        # checker = WmoChecker(param_file="./tests/WmoParameters.json")
        checker = GeneralChecks(SimpleLookupTable("./tests/test_parameters.json"))

        for message in Grib(
            "./tests/dgov-data/od_eefo_taes_sfc_2024_0001_reduced_gg.grib2"
        ):
            checker.validate(message)
